"""Some definitions of datatype used in the module."""
from typing import Tuple, Dict, Union, Any

# The record queried from the airtable
Fields = Dict[str, Any]
Record = Dict[str, Union[str, Fields]]
# The key value pair of the data document
Pair = Tuple[str, dict]



from collections import OrderedDict
from pathlib import Path
from typing import Union

import pytest

from aircopy.database import DataBase

DB = {
    'createdTime': '2019-03-04T15:42:21.000Z',
    'fields': {'Blogroll': ['recKm29VrzmM6eyjs'],
               'Collaborators': [{'createdTime': '2019-03-02T15:40:28.000Z',
                                  'fields': {'First Name': 'Randy',
                                             'ID': 'rnangah',
                                             'Institutions': [{'createdTime': '2019-03-04T16:00:32.000Z',
                                                               'fields': {'Address': 'Yaounde, Cameroon',
                                                                          'Collaborators 2': ['rec6VLeAUdItvIqbY'],
                                                                          'Name': 'University of Yaounde'},
                                                               'id': 'recFoBBdAy3cLxwYC'}],
                                             'Last Name': 'Nangah',
                                             'Manuscripts': ['recoZh0EL9qM0ufyD'],
                                             'Projects': ['rec6UVonazWUS1h94']},
                                  'id': 'rec6VLeAUdItvIqbY'}],
               'Grant': ['dmrefcheme16'],
               'Link to Analysis': 'http://gitlab.thebillingegroup.com/analysis/19st_sponge',
               'Link to Paper': 'http://gitlab.thebillingegroup.com/papers/19st_sponge',
               'Manuscripts': ['recoZh0EL9qM0ufyD'],
               'Name': '19st_sponge',
               'Notes': 'I am from Cameroon, currently working as a research '
                        'assistant for the Ministry of Scientific Research and '
                        'Innovation, and doing my PhD at the Department of '
                        'Inorganic Chemistry, University of Yaounde I, Yaounde, '
                        'Cameroon. ',
               'Project Lead': 'Songsheng Tao',
               'Samples': ['rec0Rnl9YH7h8aw33',
                           'recCOTctyb1eJXqG3',
                           'recAC9R3jPzuEn1gJ',
                           'recd8drZvs0yfRUCt'],
               'Start Date': '2019-03-04',
               'Status': '6 Report Sent'},
    'id': 'rec6UVonazWUS1h94'
}

PROJECT = OrderedDict(
    [
        ('begin_date', '2019-03-04'),
        ('collaborators', ['rnangah']),
        ('description', 'I am from Cameroon, currently working as a research '
                        'assistant for the Ministry of Scientific Research and '
                        'Innovation, and doing my PhD at the Department of '
                        'Inorganic Chemistry, University of Yaounde I, Yaounde, '
                        'Cameroon. '),
        ('grants', ['dmrefcheme16']),
        ('group_members', ['sstao']),
        ('lead', 'sstao'),
        ('log_url', ''),
        ('ana_repo_url',
         'http://gitlab.thebillingegroup.com/analysis/19st_sponge'),
        ('man_repo_url',
         'http://gitlab.thebillingegroup.com/papers/19st_sponge'),
        ('milestones',
         [
             OrderedDict([('audience',
                           ['pi', 'lead', 'group members', 'collaborators']),
                          ('due_date', '2019-03-11'),
                          ('name', 'Kick off meeting'),
                          ('objective', 'roll out of project to team'),
                          ('status', 'proposed')]),
             OrderedDict([('audience', ['pi', 'lead', 'group members']),
                          ('due_date', '2019-03-18'),
                          ('name', 'Project lead presentation'),
                          ('objective',
                           'lead presents background reading and initial '
                           'project plan'),
                          ('status', 'proposed')]),
             OrderedDict([('audience', ['pi', 'lead', 'group members']),
                          ('due_date', '2019-04-01'),
                          ('name', 'planning meeting'),
                          ('objective', 'develop a detailed plan with dates'),
                          ('status', 'proposed')]),
             OrderedDict([('audience',
                           ['pi', 'lead', 'group members', 'collaborators']),
                          ('due_date', '2020-03-03'),
                          ('name', 'submission'),
                          ('objective',
                           'submit the paper, release the code, whatever'),
                          ('status', 'proposed')])
         ]
         ),
        ('name', ''),
        ('pi_id', 'sbillinge'),
        ('status', '6 Report Sent')
    ]
)


@pytest.fixture
def fake_db() -> dict:
    return DB


@pytest.fixture
def example_project() -> OrderedDict:
    return PROJECT


LOCAL_TOKEN_FILE = Path(__file__).parent.joinpath('token.json')


def get_real_db_config(token_file: Path = LOCAL_TOKEN_FILE) -> Union[None, dict]:
    if token_file.exists():
        import json
        with token_file.open('r') as f:
            info = json.load(f)
        return info
    return None


@pytest.fixture
def real_db_config() -> Union[None, dict]:
    return get_real_db_config()


@pytest.fixture
def real_db() -> Union[None, DataBase]:
    db_config = get_real_db_config()
    if db_config is not None:
        return DataBase(db_config['base_id'], db_config['tables'], api_token=db_config['api_token'])
    return None



from typing import Any

import yaml


def write_docs_to_yaml(filename: str, content: Any) -> None:
    """Write out the documents to yaml file.

    Parameters
    ----------
    filename : str
        The path to the yaml file.

    content : Any
        The data that can be dumped by safe_dump.
    """
    with open(filename, 'w') as f:
        yaml.safe_dump(content, f)
    return



"""Parsers that parse the information queried from airtable to a pyhton dictionary following Billinge group
database standard."""
import uuid
from typing import List, Any, Tuple

import aircopy.tools as tools
from aircopy.datatype import Record, Pair

DEFAULT = {
    'pi_id': 'sbillinge',
    'name': '',
    'log_url': ''
}


def _retrieve(add_info: dict, key: str) -> Any:
    """Retrieve information from the dictionary. If not found, use the value in DEFAULT.

    Parameters
    ----------
    add_info : dict
        The addition information provided by the user.

    key : str
        The key to search inside the addition information.

    Returns
    -------
    value : Any
        The value from the add_info or DEFAULT or it is None.
    """
    return add_info.get(key, DEFAULT.get(key))


def parse_project(record: Record, add_info: dict) -> Tuple[Pair, List[Pair], List[Pair]]:
    """Parse the project record in airtable to Billinge group format.

    Return a key-value pair of the project and a list of key-value pairs of the people doc and institution doc
    in the project. The record should be denormalized at first.

    Parameters
    ----------
    record : Record
        The record from the airtable.

    add_info : dict
        A dictionary of the additional information.

    Returns
    -------
    project : tuple
        The key-value pair of project document.

    people : list
        The list of the key-value pairs of the people in the collaborators list.

    institutions : list
        The list of the key-value pairs of the institutions of those collaborators.
    """
    record = tools.get_data(record)
    pairs = list(map(parse_person, record.get('Collaborators', [])))
    people = [pair[0] for pair in pairs if pair[0] is not None]
    institutions = [pair[1] for pair in pairs if pair[1] is not None]
    key = record.get('Name')
    value = {
        'begin_date': record.get('Start Date'),
        'collaborators': tools.get_keys(people),
        'description': record.get('Notes'),
        'grants': record.get('Grant'),
        'group_members': [tools.gen_person_id(record.get('Project Lead'))],
        'lead': tools.gen_person_id(record.get('Project Lead')),
        'log_url': _retrieve(add_info, 'log_url'),
        'ana_repo_url': record.get('Link to Analysis'),
        'man_repo_url': record.get('Link to Paper'),
        'milestones': tools.auto_gen_milestons(record.get('Start Date')),
        'name': _retrieve(add_info, 'name'),
        'pi_id': _retrieve(add_info, 'pi_id'),
        'status': record.get('Status')
    }
    project = (key, value)
    return project, people, institutions


def parse_person(record: Record) -> Tuple[Pair, Pair]:
    """Parse the person information in the People airtable to the contact and institution document.

    Parameters
    ----------
    record : Record
        The record from the airtable.

    Returns
    -------
    contact : tuple
        The key-value pair of the contact document.

    institution : tuple
        The key-value pair of the institution document.
    """
    record = tools.get_data(record)
    institutions = list(map(parse_institution, record.get('Institutions', [])))
    institution = institutions[0] if institutions else None
    key = record.get('ID')
    value = {
        'name': '{} {}'.format(
            record.get('First Name'),
            record.get('Last Name')
        ),
        'aka': '{}, {}'.format(
            record.get('Last Name'),
            record.get('First Name')
        ),
        'institution': institution[0] if institution else None,
        'notes': [],
        'uuid': str(uuid.uuid4())
    }
    tools.tag_date(value)
    contact = (key, value)
    return contact, institution


def parse_institution(record: Record) -> Tuple[str, dict]:
    """Parse the record from the Institutions airtbale to the institution documents.

    Parameters
    ----------
    record : Record
        The record from the airtable.

    Returns
    -------
    institution : tuple
        The key-value pair of the institution document.
    """
    record = tools.get_data(record)
    key = tools.gen_inst_id(record.get('Name', ''), mode='auto')
    value = {
        'aka': [],
        'city': record.get('Address'),
        'state': record.get('State'),
        'county': record.get('County'),
        'name': record.get('Name')
    }
    tools.tag_date(value)
    institution = (key, value)
    return institution



import aircopy.parser as parser


def test_parse_institution(fake_db):
    record = fake_db.get('fields').get('Collaborators')[0].get('fields').get('Institutions')[0]
    key, value = parser.parse_institution(record)
    assert key == 'uyaounde'
    assert value['name'] == 'University of Yaounde'
    assert value['aka'] == []
    assert value['city'] == 'Yaounde, Cameroon'
    assert value['state'] is None
    assert value['county'] is None


def test_parse_person(fake_db):
    record = fake_db.get('fields').get('Collaborators')[0]
    (key, value), _ = parser.parse_person(record)
    assert key == 'rnangah'
    assert value['aka'] == 'Nangah, Randy'
    assert value['institution'] == 'uyaounde'
    assert value['name'] == 'Randy Nangah'
    assert value['notes'] == []


def test_parse_project(fake_db, example_project):
    record = fake_db
    (key, value), _, _ = parser.parse_project(record, {})
    assert key == '19st_sponge'
    assert value == example_project



import aircopy.tools as tools


def test_gen_person_id():
    assert tools.gen_person_id('Hahaha LaLaLa') == 'hlalala'
    assert tools.gen_person_id('Songsheng Tao') == 'sstao'


def test_auto_gen_milestons():
    milestones = tools.auto_gen_milestons('2020-01-01')
    assert milestones[0]['due_date'] == '2020-01-08'
    assert milestones[1]['due_date'] == '2020-01-15'
    assert milestones[2]['due_date'] == '2020-01-29'
    assert milestones[3]['due_date'] == '2020-12-31'


def test_tag_date():
    record = dict()
    tools.tag_date(record)
    assert set(record.keys()) == {'day', 'month', 'year', 'updated'}


def test_get_keys():
    assert tools.get_keys([('k', 'v'), ('k', 'v')]) == ['k', 'k']
    assert tools.get_keys([]) == []


def test_gen_inst_id():
    assert tools.gen_inst_id('University of Hahaha', 'auto') == 'uhahaha'
    assert tools.gen_inst_id('Department of Hahaha, University of Hahaha', 'auto') == 'hahaha,universityofhahaha'
    assert tools.gen_inst_id('School of Hahaha, University of Hahaha', 'auto') == 'hahaha,universityofhahaha'


def test_get_data(fake_db):
    assert isinstance(tools.get_data(fake_db), dict)



"""The cli interface."""
import os
from typing import Union, List

import fire

import aircopy.io as io
from aircopy.database import DataBase


def aircopy(add_info: dict = None, max_records: int = None, view: int = None, page_size: int = None,
            fields: Union[str, List[str]] = None, sort: List[str] = None, formula: str = None):
    """Transfer the data from Airtbale to Billinge group projecta database.

    It will read the variables 'AIRTABLE_BASE_ID' and 'AIRTABLE_API_TOKEN' inside your environment and connect
    to your airtable database, download the content and parse it into Billinge group database documents and
    write it out to yaml files. The yaml files will be named as "airtable_projecta.yaml", "airtable_contacts.yaml"
    and "airtable_institutions.yaml".

    To use the it, you need to find the base id of your Airtable by visiting the airtable api website (
    https://airtable.com/api) and generate the api token following the instructions on the website (
    https://support.airtable.com/hc/en-us/articles/219046777-How-do-I-get-my-API-key-). Then, export the base id
    and api token to the environment variables.

    $ export AIRTABLE_BASE_ID=***

    $ export AIRTABLE_API_TOKEN=***

    Or you can add these lines in your ~/.bashrc so that you don't need to repeat doing it. Then you can run the
    aircopy.

    $ aircopy

    Aircopy supports the data query options provided by the airtable. For example, if you would like to transfer
    the project with project lead "Songsheng Tao", then you need to add the formula.

    $ aircopy --formula "{Project Lead} = 'Songsheng Tao'"

    For more information, please read the https://airtable.com/api to see how to use the flags.

    Also, you can specify what additional information you would like to include in the projecta docs by passing
    a dictionary to the 'add_info' flag. The keys should be inside the Billinge group projecta schemas. For
    example, you would like your pi_id to be someone else like 'sstao' (default 'sbillinge'), you can use

    $ aircopy --add_info {pi_id:sstao}

    Parameters
    ----------
    add_info : dict
        Additional information for the projecta. The keys must be in the Billinge group
        projecta database schemas.

    max_records : int
        The maximum total number of records that will be returned.

    view : str
        The name or ID of a view.

    page_size : int
        The number of records returned in each request. Must be less than or equal to 100. Default is 100.

    fields : str or list
        Name of field or fields to be retrieved. Default is all fields.

    sort : list
        List of fields to sort by. Default order is ascending.

    formula : str
        Airtable formula.
    """
    if add_info is None:
        add_info = dict()
    options = dict()
    if max_records is not None:
        options.update({'add_info': max_records})
    if view is not None:
        options.update({'view': view})
    if page_size is not None:
        options.update({'page_size': page_size})
    if fields is not None:
        options.update({'fields': fields})
    if sort is not None:
        options.update({'sort': sort})
    if formula is not None:
        options.update({'formula': formula})
    base_id = os.getenv('AIRTABLE_BASE_ID')
    if base_id is None:
        raise EnvironmentError(
            "AIRTABLE_BASE_ID not found in environment. "
            "Please add the environment variable AIRTABLE_BASE_ID"
        )
    api_token = os.getenv('AIRTABLE_API_TOKEN')
    if api_token is None:
        raise EnvironmentError(
            "AIRTABLE_API_TOKEN not found in environment. "
            "Please add the environment variable AIRTABLE_API_TOKEN"
        )
    db = DataBase(base_id, api_token=api_token)
    projects, people, institutions = db.get_projecta(add_info, **options)
    io.write_docs_to_yaml('airtable_projects.yaml', projects)
    io.write_docs_to_yaml('airtable_people.yaml', people)
    io.write_docs_to_yaml('airtable_institutions.yaml', institutions)
    return


def main():
    """Fire aircopy."""
    fire.Fire(aircopy)



"""The tools for data operation and the wrapper for the airtable database."""
import copy
from typing import Union, List, Tuple, Generator

from airtable import Airtable

import aircopy.parser as parser
import aircopy.tools as tools
from aircopy.datatype import Record


def query(db: Airtable, uid: str, only_data: bool):
    """Query the data by the unique id from airtable.

    Parameters
    ----------
    db : Airtable
        The airtable client.

    uid : str
        The airtbale style uid.

    only_data : bool
        If True, only the value for the 'fields' in record will be returned. Otherwise, return whole record.

    Returns
    -------
    record : Record or Fields
        The dictionary of the data.
    """
    record = db.get(uid)
    if only_data:
        return record.get('fields')
    return record


def denormalize_project(project: Record, people: Airtable, institutions: Airtable, only_data: bool = False,
                        inplace: bool = False) -> Union[None, dict]:
    """Denormalize the project record from airtbale.

    Parameters
    ----------
    project : Record
        The record the project.

    people : Airtable
        The airtable client of 'People'.

    institutions : Airtable
        The airtable client of 'Institutions'.

    only_data : bool
        If True, only the value for the 'fields' in record will be returned. Otherwise, return whole record.

    inplace : bool
        If True, demoralization is done inplace. Otherwise, return the denormalized copy of the project record.

    Returns
    -------
    project : None or dict
        If inplace == False, return the denormalized record. Otherwise, denormalize inplace and return None.
    """
    if not inplace:
        project = copy.deepcopy(project)
    project = tools.get_data(project)
    if 'Collaborators' in project:
        project['Collaborators'] = [
            query(people, uid, only_data=only_data)
            for uid in project['Collaborators']
        ]
        for collaborator in project['Collaborators']:
            fields = tools.get_data(collaborator) if not only_data else collaborator
            if 'Institutions' not in fields:
                fields['Institutions'] = []
            else:
                fields['Institutions'] = [
                    query(institutions, uid, only_data=only_data)
                    for uid in fields['Institutions']
                ]
    if not inplace:
        return project
    return


def get_projecta_docs(projects: Airtable, people: Airtable, institutions: Airtable, add_info: dict,
                      **options) -> Generator:
    """Generate the projecta documents and the documents of contacts and institutions from airtbale database.

    Parameters
    ----------
    projects : Airtable
        The Projects database.

    people : Airtable
        The People database.

    institutions : Airtable
        The institutions database.

    add_info : dict
        A dictionary of the additional information.

    options : dict
        The view and filter options. Include
            max_records (``int``, optional): The maximum total number of
                records that will be returned. See :any:`MaxRecordsParam`
            view (``str``, optional): The name or ID of a view.
                See :any:`ViewParam`.
            page_size (``int``, optional ): The number of records returned
                in each request. Must be less than or equal to 100.
                Default is 100. See :any:`PageSizeParam`.
            fields (``str``, ``list``, optional): Name of field or fields to
                be retrieved. Default is all fields. See :any:`FieldsParam`.
            sort (``list``, optional): List of fields to sort by.
                Default order is ascending. See :any:`SortParam`.
            formula (``str``, optional): Airtable formula.
                See :any:`FormulaParam`.

    Yields
    ------
    project : tuple
        The key-value pair of project document.

    people : list
        The list of the key-value pairs of the people in the collaborators list.

    institutions : list
        The list of the key-value pairs of the institutions of those collaborators.
    """
    for page in projects.get_iter(**options):
        for record in page:
            denormalize_project(record, people, institutions, inplace=True)
            _project, _people, _institutions = parser.parse_project(record, add_info)
            yield _project, _people, _institutions


class DataBase:
    """A collection of airtables at the same base."""

    def __init__(self, base_id: str, tables: List[str] = ('Projects', 'People', 'Institutions'),
                 api_token: str = None):
        """
        Initiate the class. Attributes will be added as the name of the table and its corresponding Airtable
        object.

        Parameters
        ----------
        base_id : str
            The base id for the airtables.

        tables : list
            The name of the tables.
        """
        for table in tables:
            self.__setattr__(table, Airtable(base_id, table, api_key=api_token))

    def get_projecta(self, add_info: dict, **options) -> Tuple[dict, dict, dict]:
        """Generate the projecta documents and the documents of contacts and institutions from airtbale database.

        Parameters
        ----------
        add_info : dict
            A dictionary of the additional information.

        options : kwargs
            The view and filter options. Include
                max_records (``int``, optional): The maximum total number of
                    records that will be returned. See :any:`MaxRecordsParam`
                view (``str``, optional): The name or ID of a view.
                    See :any:`ViewParam`.
                page_size (``int``, optional ): The number of records returned
                    in each request. Must be less than or equal to 100.
                    Default is 100. See :any:`PageSizeParam`.
                fields (``str``, ``list``, optional): Name of field or fields to
                    be retrieved. Default is all fields. See :any:`FieldsParam`.
                sort (``list``, optional): List of fields to sort by.
                    Default order is ascending. See :any:`SortParam`.
                formula (``str``, optional): Airtable formula.
                    See :any:`FormulaParam`.

        Returns
        -------
        project : dict
            The dictionary of project documents.

        people : dict
            The dictionary of the people in the collaborators list.

        institutions : dict
            The dictionary of the institutions of those collaborators.
        """
        gen = get_projecta_docs(
            getattr(self, 'Projects'), getattr(self, 'People'), getattr(self, 'Institutions'), add_info, **options
        )
        projects, people, institutions = dict(), dict(), dict()
        for _project, _people, _institutions in gen:
            projects.update([_project])
            people.update(_people)
            institutions.update(_institutions)
        return projects, people, institutions



import os
from tempfile import TemporaryDirectory

import aircopy.cli as cli


def test_aircopy(real_db_config):
    if real_db_config is not None:
        with TemporaryDirectory() as d:
            os.chdir(d)
            os.environ['AIRTABLE_BASE_ID'] = real_db_config['base_id']
            os.environ['AIRTABLE_API_TOKEN'] = real_db_config['api_token']
            cli.aircopy(formula="{Name} = '19st_TiO2B'")



import sys
from os import path

from setuptools import setup, find_packages

import versioneer

# NOTE: This file must remain Python 2 compatible for the foreseeable future,
# to ensure that we error out properly for people with outdated setuptools
# and/or pip.
min_version = (3, 7)
if sys.version_info < min_version:
    error = """
aircopy does not support Python {0}.{1}.
Python {2}.{3} and above is required. Check your Python version like so:

python3 --version

This may be due to an out-of-date pip. Make sure you have pip >= 9.0.1.
Upgrade pip like so:

pip install --upgrade pip
""".format(*(sys.version_info[:2] + min_version))
    sys.exit(error)

here = path.abspath(path.dirname(__file__))

with open(path.join(here, 'README.rst'), encoding='utf-8') as readme_file:
    readme = readme_file.read()

with open(path.join(here, 'requirements.txt')) as requirements_file:
    # Parse requirements.txt, ignoring any commented-out lines.
    requirements = [line for line in requirements_file.read().splitlines()
                    if not line.startswith('#')]

setup(
    name='aircopy',
    version=versioneer.get_version(),
    cmdclass=versioneer.get_cmdclass(),
    description="The CLI software to copy airtable data to Billinge group database",
    long_description=readme,
    author="Songsheng Tao",
    author_email='st3107@columbia.edu',
    url='https://github.com/st3107/aircopy',
    python_requires='>={}'.format('.'.join(str(n) for n in min_version)),
    packages=find_packages(exclude=['docs', 'tests']),
    entry_points={
        'console_scripts': [
            'aircopy = aircopy.cli:main',
        ],
    },
    include_package_data=True,
    package_data={
        'aircopy': [
            # When adding files here, remember to update MANIFEST.in as well,
            # or else they will not be included in the distribution on PyPI!
            # 'path/to/data_file',
        ]
    },
    install_requires=requirements,
    license="BSD (3-clause)",
    classifiers=[
        'Development Status :: 2 - Pre-Alpha',
        'Natural Language :: English',
        'Programming Language :: Python :: 3',
    ],
)



"""Tools used in the module."""
import copy
from datetime import datetime, timedelta
from typing import List, Iterable, Union

from nameparser import HumanName

from aircopy.datatype import Record

MILESTONES_TEMPLATE = [
    {
        'audience': ['pi', 'lead', 'group members', 'collaborators'],
        'due_date': timedelta(days=7),
        'name': 'Kick off meeting',
        'objective': 'roll out of project to team',
        'status': 'proposed'
    },
    {
        'audience': ['pi', 'lead', 'group members'],
        'due_date': timedelta(days=14),
        'name': 'Project lead presentation',
        'objective': 'lead presents background reading and initial project plan',
        'status': 'proposed'
    },
    {
        'audience': ['pi', 'lead', 'group members'],
        'due_date': timedelta(days=28),
        'name': 'planning meeting',
        'objective': 'develop a detailed plan with dates',
        'status': 'proposed'
    },
    {
        'audience': ['pi', 'lead', 'group members', 'collaborators'],
        'due_date': timedelta(days=365),
        'name': 'submission',
        'objective': 'submit the paper, release the code, whatever',
        'status': 'proposed'}
]

SPECIAL_ID = {
    'Songsheng Tao': 'sstao'
}


def gen_person_id(name: str) -> Union[str, None]:
    """Generate the id of the person."""
    if name is None:
        return
    if name in SPECIAL_ID:
        return SPECIAL_ID.get(name)
    hn = HumanName(name)
    return '{}{}'.format(hn.first[0], hn.last).lower()


def auto_gen_milestons(start_date: str, template: List[dict] = None) -> List[dict]:
    """Automatically generate the milestones list according to the template."""
    if template is None:
        template = copy.deepcopy(MILESTONES_TEMPLATE)
    start_date = datetime.strptime(start_date, '%Y-%m-%d')
    for milestone in template:
        time_gap = milestone['due_date']
        due_date = start_date + time_gap
        milestone['due_date'] = due_date.strftime('%Y-%m-%d')
    return template


def tag_date(record: dict, turn_off: tuple = ()):
    """Tag the record dictionary with today date time."""
    today = datetime.today()
    now = today.now()
    date = {
        'day': today.day,
        'month': today.month,
        'year': today.year,
        'updated': now
    }
    for key in turn_off:
        date.pop(key)
    record.update(date)
    return


def get_keys(pairs: Iterable[tuple]) -> list:
    """Get the key of a iterable of key value pairs."""
    return [pair[0] for pair in pairs]


def gen_inst_id(name: str, mode: str):
    """Generate the key according to the name. Four mode u, d, s, auto."""

    def gen_key_u(_name: str):
        return str(_name.lower().replace(' of ', "").replace(" ", "").replace("university", "u"))

    def gen_key_d(_name: str):
        _name = _name.lower()
        if "department" in _name:
            return _name.replace("department of", "").replace("department", "").replace(" ", "")
        else:
            return ''.join(_name.split())

    def gen_key_s(_name: str):
        _name = _name.lower()
        if "school" in _name:
            return _name.replace("school of", "").replace("school", "").replace(" ", "")
        else:
            return ''.join(_name.split())

    def auto_mode(_name: str):
        _name = _name.lower()
        if "school" in _name:
            return 's'
        if "department" in _name:
            return 'd'
        return 'u'

    dct = {
        'u': gen_key_u,
        'd': gen_key_d,
        's': gen_key_s
    }
    if mode not in dct and mode != 'auto':
        raise ValueError(f"Unknown mode: {mode}")
    if mode == 'auto':
        mode = auto_mode(name)
    method = dct.get(mode)
    return method(name)


def get_data(record: Record):
    """Get the data in a record from the airtable api."""
    if 'fields' not in record:
        raise ValueError(
            "'fields' not found in the following record {}".format(record)
        )
    return record.get('fields')



import aircopy.database as database


def test_denormalize_project(real_db):
    if real_db is None:
        return
    record = real_db.Projects.get('rec6UVonazWUS1h94')
    people = getattr(real_db, 'People')
    institutions = getattr(real_db, 'Institutions')
    assert database.denormalize_project(record, people, institutions, inplace=False)
    assert database.denormalize_project(record, people, institutions, inplace=False, only_data=True)


def test_get_projecta(real_db):
    if real_db is None:
        return
    assert real_db.get_projecta({}, formula="{Name} = '19st_TiO2B'")



import tempfile
from pathlib import Path

import aircopy.io as io


def test_write_docs_to_yaml():
    dct = {'k1': 'v1', 'k2': 'v2'}
    with tempfile.TemporaryDirectory() as d:
        filename = Path(d).joinpath('test.yaml')
        io.write_docs_to_yaml(str(filename), dct)
        with filename.open('r') as f:
            assert f.read() == "k1: v1\nk2: v2\n"
